package ly.hall.rose.healthsync;

import com.mojang.authlib.GameProfile;
import com.mojang.brigadier.Command;
import com.mojang.brigadier.CommandDispatcher;
import com.mojang.brigadier.builder.LiteralArgumentBuilder;
import com.mojang.brigadier.context.CommandContext;
import com.mojang.brigadier.exceptions.CommandSyntaxException;
import net.fabricmc.api.ModInitializer;

import net.fabricmc.fabric.api.command.v2.CommandRegistrationCallback;
import net.fabricmc.fabric.api.entity.event.v1.ServerLivingEntityEvents;
import net.fabricmc.fabric.api.entity.event.v1.ServerPlayerEvents;
import net.fabricmc.fabric.api.networking.v1.PacketByteBufs;
import net.fabricmc.fabric.api.networking.v1.ServerPlayConnectionEvents;
import net.fabricmc.fabric.api.networking.v1.ServerPlayNetworking;
import net.minecraft.command.CommandRegistryAccess;
import net.minecraft.command.CommandSource;
import net.minecraft.entity.LivingEntity;
import net.minecraft.entity.attribute.EntityAttributes;
import net.minecraft.entity.damage.DamageSource;
import net.minecraft.entity.effect.StatusEffects;
import net.minecraft.nbt.NbtCompound;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.scoreboard.ScoreboardCriterion;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.command.CommandManager;
import net.minecraft.server.command.ServerCommandSource;
import net.minecraft.server.network.ServerPlayerEntity;
import net.minecraft.text.Text;
import net.minecraft.world.GameMode;
import net.minecraft.world.PersistentState;
import net.minecraft.world.PersistentStateManager;
import net.minecraft.world.World;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.mojang.brigadier.builder.LiteralArgumentBuilder.literal;

class PlayerStats {
	public float TotalDamageTaken;
	public float TotalHealed;
}

record TaggedPlayerCoarseStats(String tag, int TotalHealthDamageTaken, int TotalHealthHealed) {}

class GlobalHealthSaverLoader extends PersistentState {
	public static GlobalHealthSaverLoader getServerState(MinecraftServer server) {
		PersistentStateManager psm = server.getWorld(World.OVERWORLD).getPersistentStateManager();
		GlobalHealthSaverLoader state = psm.getOrCreate(GlobalHealthSaverLoader::createFromNbt, GlobalHealthSaverLoader::new, "healthsync");
		state.markDirty();

		return state;
	}

	public static PlayerStats getPlayerStats(LivingEntity player) {
		var serverState = getServerState(Objects.requireNonNull(player.getWorld().getServer()));

		return serverState.players.computeIfAbsent(player.getUuid(), uuid -> new PlayerStats());
	}

	public static Stream<Map.Entry<UUID, PlayerStats>> getAllPlayerStats(MinecraftServer server) {
		return getServerState(server).players.entrySet().stream();
	}

	public float GlobalHealth = 20;
	public float GlobalAbsorption = 0;

	public HashMap<UUID, PlayerStats> players = new HashMap<>();

	public static GlobalHealthSaverLoader createFromNbt(NbtCompound tag) {
		var s = new GlobalHealthSaverLoader();
		s.GlobalHealth = tag.getFloat("healthsync:globalHealth");
		s.GlobalAbsorption = tag.getFloat("healthsync:globalAbsorption");
		NbtCompound playersNbt = tag.getCompound("healthsync:players");
		playersNbt.getKeys().forEach(key -> {
			PlayerStats playerData = new PlayerStats();

			playerData.TotalHealed = playersNbt.getCompound(key).getFloat("totalhealed");
			playerData.TotalDamageTaken = playersNbt.getCompound(key).getFloat("totaldamagetaken");

			UUID uuid = UUID.fromString(key);
			s.players.put(uuid, playerData);
		});
		return s;
	}

	public NbtCompound writeNbt(NbtCompound nbt) {
		nbt.putFloat("healthsync:globalHealth", GlobalHealth);
		nbt.putFloat("healthsync:globalAbsorption", GlobalAbsorption);
		NbtCompound playersNbt = new NbtCompound();
		players.forEach((uuid, playerData) -> {
			NbtCompound playerNbt = new NbtCompound();

			playerNbt.putFloat("totalhealed", playerData.TotalHealed);
			playerNbt.putFloat("totaldamagetaken", playerData.TotalDamageTaken);

			playersNbt.put(uuid.toString(), playerNbt);
		});
		nbt.put("healthsync:players", playersNbt);
		return nbt;
	}
}

public class HealthSync implements ModInitializer {
	// This logger is used to write text to the console and the log file.
	// It is considered best practice to use your mod id as the logger's name.
	// That way, it's clear which mod wrote info, warnings, and errors.
    public static final Logger LOGGER = LoggerFactory.getLogger("healthsync");

	@Override
	public void onInitialize() {
		var healthHealedCriterion = ScoreboardCriterion.create("healthsync_health_contributed", true, ScoreboardCriterion.RenderType.INTEGER);
		var damageTakenCriterion = ScoreboardCriterion.create("healthsync_damage_contributed", true, ScoreboardCriterion.RenderType.INTEGER);

		CommandRegistrationCallback.EVENT.register((CommandDispatcher<ServerCommandSource> dispatcher, CommandRegistryAccess registryAccess, CommandManager.RegistrationEnvironment environment) -> {
			LiteralArgumentBuilder<ServerCommandSource> builder = literal("healthsync");
			LiteralArgumentBuilder<ServerCommandSource> builder2 = literal("scoreboard");

			dispatcher.register(builder.then(builder2.executes(c -> {
				try {
					var source = c.getSource();

					var stats = GlobalHealthSaverLoader.getAllPlayerStats(source.getServer())
							.map(v -> new TaggedPlayerCoarseStats(
									Objects.requireNonNull(source.getServer().getUserCache()).getByUuid(v.getKey()).map(GameProfile::getName).orElse(v.getKey().toString()),
									(int) Math.ceil(v.getValue().TotalDamageTaken),
									(int) Math.ceil(v.getValue().TotalHealed)
							))
							.sorted(Comparator.comparingInt(TaggedPlayerCoarseStats::TotalHealthDamageTaken))
							.toList();

					int maxNameLength = 0;
					int maxHealLength = 0;
					int maxDamageLength = 0;

					for (TaggedPlayerCoarseStats stat : stats) {
						maxNameLength = Math.max(maxNameLength, stat.tag().length());
						maxHealLength = Math.max(maxHealLength, Float.toString(stat.TotalHealthHealed() / 2.0F).length());
						maxDamageLength = Math.max(maxDamageLength, Float.toString(stat.TotalHealthDamageTaken() / 2.0F).length());
					}

					for (TaggedPlayerCoarseStats stat : stats) {
						source.sendMessage(Text.of(StringUtils.rightPad(stat.tag(), maxNameLength) + " | " + StringUtils.leftPad(Float.toString(stat.TotalHealthHealed() / 2.0F), maxHealLength) + "\uD83D\uDDA4 | " + StringUtils.leftPad(Float.toString(stat.TotalHealthDamageTaken() / 2.0F), maxDamageLength) + "\uD83D\uDC94"));
					}

					return 1;
				} catch (Exception e) {
					HealthSync.LOGGER.error(e.toString());

					return 1;
				}
			})));
		});

		LivingEntityDamageHandler.initializeGeneric();

		LivingEntityHealHandler.handle((e, reason, deltaHealth, deltaAbsorption) -> {
			var state = GlobalHealthSaverLoader.getServerState(Objects.requireNonNull(e.getServer()));

			state.GlobalHealth += deltaHealth;
			state.GlobalAbsorption += deltaAbsorption;

			if (state.GlobalAbsorption < 0)
				state.GlobalAbsorption = 0;

			var stats = GlobalHealthSaverLoader.getPlayerStats(e);

			stats.TotalHealed += deltaHealth + deltaAbsorption;

			e.getScoreboard().forEachScore(healthHealedCriterion, e.getEntityName(), innerScore -> innerScore.setScore((int) stats.TotalHealed));

			synchronizeHealth(Objects.requireNonNull(e.getServer()));

			for (ServerPlayerEntity serverPlayerEntity : e.getServer().getPlayerManager().getPlayerList()) {
				var buf = PacketByteBufs.create();

				buf.writeUuid(e.getUuid());

				buf.writeOptional(Optional.ofNullable(reason), (bufx, pos) -> {
                    assert reason != null;
                    bufx.writeString(reason.toString());
				});

				buf.writeFloat(deltaHealth);
				buf.writeFloat(deltaAbsorption);

				ServerPlayNetworking.send(serverPlayerEntity, NetworkingConstants.HEAL_EVENT_PACKET_ID, buf);
			}
		});

		LivingEntityDamageHandler.handle((e, source, deltaHealth, deltaAbsorption) -> {
			var state = GlobalHealthSaverLoader.getServerState(Objects.requireNonNull(e.getServer()));

			state.GlobalHealth += deltaHealth;
			state.GlobalAbsorption += deltaAbsorption;

			if (state.GlobalAbsorption < 0)
				state.GlobalAbsorption = 0;

			var stats = GlobalHealthSaverLoader.getPlayerStats(e);

			stats.TotalDamageTaken += (-1 * deltaHealth) + (-1 * deltaAbsorption);

			e.getScoreboard().forEachScore(damageTakenCriterion, e.getEntityName(), innerScore -> innerScore.setScore((int) stats.TotalDamageTaken));

			synchronizeHealth(Objects.requireNonNull(e.getServer()));

			for (ServerPlayerEntity serverPlayerEntity : e.getServer().getPlayerManager().getPlayerList()) {
				var buf = PacketByteBufs.create();

				buf.writeUuid(e.getUuid());

				buf.writeOptional(Optional.ofNullable(source), (bufx, pos) -> {
                    assert source != null;

                    int sourceTypeId = e.getWorld().getRegistryManager().get(RegistryKeys.DAMAGE_TYPE).getRawId(source.getType());
					int attackerId = source.getAttacker() == null ? 0 : (source.getAttacker().getId() + 1);
					int sourceId = source.getSource() == null ? 0 : (source.getSource().getId() + 1);

					bufx.writeInt(sourceTypeId);
					bufx.writeInt(attackerId);
					bufx.writeInt(sourceId);

					bufx.writeOptional(Optional.ofNullable(source.getStoredPosition()), (bufxx, poss) -> {
						bufxx.writeDouble(source.getStoredPosition().getX());
						bufxx.writeDouble(source.getStoredPosition().getY());
						bufxx.writeDouble(source.getStoredPosition().getZ());
					});
				});

				buf.writeFloat(deltaHealth);
				buf.writeFloat(deltaAbsorption);
				buf.writeBoolean(e.hasStatusEffect(StatusEffects.POISON));
				buf.writeBoolean(e.hasStatusEffect(StatusEffects.WITHER));
				buf.writeBoolean(e.isFrozen());

				ServerPlayNetworking.send(serverPlayerEntity, NetworkingConstants.DAMAGE_EVENT_PACKET_ID, buf);
			}
		});

		Object o = new Object();

		ServerPlayConnectionEvents.JOIN.register((handler, sender, server) -> {
			synchronized(o) {
				var state = GlobalHealthSaverLoader.getServerState(server);

				LivingEntityHealHandler.applyReason(LivingEntityHealReason.Synchronization);
				handler.player.setHealth(state.GlobalHealth);

				LivingEntityHealHandler.applyReason(LivingEntityHealReason.Synchronization);
				handler.player.setAbsorptionAmount(state.GlobalAbsorption);
			}
		});

		ServerPlayerEvents.AFTER_RESPAWN.register((oldPlayer, newPlayer, alive) -> {
			if (newPlayer.getWorld().getPlayers().stream().allMatch(e -> e.isDead() || e.getUuid().equals(newPlayer.getUuid()))) {
				var state = GlobalHealthSaverLoader.getServerState(Objects.requireNonNull(newPlayer.getServer()));

				state.GlobalHealth = 20;
				state.GlobalAbsorption = 0;
			}
		});
	}

	private void synchronizeHealth(MinecraftServer server) {
		for (ServerPlayerEntity serverPlayerEntity : server.getPlayerManager().getPlayerList()) {
			var state = GlobalHealthSaverLoader.getServerState(Objects.requireNonNull(server));

			if (serverPlayerEntity.getHealth() < state.GlobalHealth) {
				LivingEntityHealHandler.applyReason(LivingEntityHealReason.Synchronization);
				serverPlayerEntity.setHealth(state.GlobalHealth);
			}

			if (serverPlayerEntity.getHealth() > state.GlobalHealth) {
				serverPlayerEntity.setHealth(state.GlobalHealth);
			}

			if (serverPlayerEntity.getAbsorptionAmount() < state.GlobalAbsorption) {
				LivingEntityHealHandler.applyReason(LivingEntityHealReason.Synchronization);
				serverPlayerEntity.setAbsorptionAmount(state.GlobalAbsorption);
			}

			if (serverPlayerEntity.getAbsorptionAmount() > state.GlobalAbsorption) {
				serverPlayerEntity.setAbsorptionAmount(state.GlobalAbsorption);
			}
		}
	}
}